import torch


def negative_part(x):
    return torch.minimum(x, torch.zeros_like(x))


def positive_part(x):
    return torch.maximum(x, torch.zeros_like(x))


def as_channel_parameters(parameter: torch.Tensor, x: torch.Tensor):
    shape = [1] * x.dim()
    shape[1] = -1
    return parameter.view(*shape)


def device_compatibility_check(model, x: torch.Tensor):
    for p in model.parameters():
        if p.device != x.device or p.dtype != x.dtype:
            p.data = p.data.to(device=x.device, dtype=x.dtype)


def phi_hat(x, a, lam):
    term_pos = torch.maximum(lam - torch.abs(x - a), torch.zeros_like(x))
    term_neg = torch.minimum(torch.abs(x - (a + 2 * lam)) - lam, torch.zeros_like(x))
    return term_pos + term_neg



import torch
from torch import nn
from torch.nn.parameter import Parameter, UninitializedParameter
from torch.nn.modules.lazy import LazyModuleMixin
import torch.nn.functional as F


class PDELU(LazyModuleMixin, nn.Module):
    def __init__(self, theta: float = 0.5):
        super().__init__()
        if theta == 1.0:
            raise ValueError(
                "theta cannot be 1.0, as it would cause a division by zero."
            )
        self.theta = float(theta)
        self._power_val = 1.0 / (1.0 - self.theta)
        self.alpha: torch.Tensor = UninitializedParameter()
        self._num_channels = None

    def _infer_parameters(self, x: torch.Tensor):
        if x.ndim < 2:
            raise ValueError(
                f"Input tensor must have at least 2 dimensions (N, C), but got shape {tuple(x.shape)}"
            )
        c = int(x.shape[1])
        self._num_channels = c
        param_shape = [1] * x.ndim
        param_shape[1] = c

        with torch.no_grad():
            self.alpha = Parameter(
                self.alpha.new_empty(param_shape, dtype=x.dtype, device=x.device).fill_(
                    0.1
                )
            )
        self._lazy_materialized = True

    def reset_parameters(self):
        if not isinstance(self.alpha, UninitializedParameter):
            with torch.no_grad():
                self.alpha.fill_(0.1)

    def forward(self, x: torch.Tensor):
        if isinstance(self.alpha, UninitializedParameter):
            self._infer_parameters(x)

        if x.ndim < 2:
            raise ValueError(
                f"Input tensor must have at least 2 dimensions (N, C), but got shape {tuple(x.shape)}"
            )
        if self._num_channels is not None and x.shape[1] != self._num_channels:
            raise RuntimeError(
                f"PDELU was initialized with C={self._num_channels} but got C={x.shape[1]}. "
                "Create a new PDELU for a different channel size."
            )
        positive_part = F.relu(x)
        inner_term = F.relu(1.0 + (1.0 - self.theta) * x)
        powered_term = torch.pow(inner_term, self._power_val)
        subtracted_term = powered_term - 1.0
        zero = torch.zeros(1, dtype=x.dtype, device=x.device)
        negative_part = self.alpha * torch.minimum(subtracted_term, zero)
        return positive_part + negative_part



import torch
from torch import nn
from torch.nn.parameter import Parameter, UninitializedParameter
from torch.nn.modules.lazy import LazyModuleMixin
import torch.nn.functional as F


class WideMELU(LazyModuleMixin, nn.Module):
    def __init__(self, maxInput: float = 1.0):
        super().__init__()
        self.maxInput = float(maxInput)
        self.alpha: torch.Tensor = UninitializedParameter()
        self.beta: torch.Tensor = UninitializedParameter()
        self.gamma: torch.Tensor = UninitializedParameter()
        self.delta: torch.Tensor = UninitializedParameter()
        self.xi: torch.Tensor = UninitializedParameter()
        self.psi: torch.Tensor = UninitializedParameter()
        self.theta: torch.Tensor = UninitializedParameter()
        self.lam: torch.Tensor = UninitializedParameter()
        self.num_channels = None

    def _infer_parameters(self, x: torch.Tensor):
        if x.dim() != 4:
            raise ValueError(
                f"Expected 4D input (N, C, H, W), got {x.dim()}D with shape {tuple(x.shape)}"
            )

        c = int(x.shape[1])
        self.num_channels = c
        shape = (1, c, 1, 1)

        with torch.no_grad():
            self.alpha = Parameter(
                self.alpha.new_empty(shape, dtype=x.dtype, device=x.device).zero_()
            )
            self.beta = Parameter(
                self.beta.new_empty(shape, dtype=x.dtype, device=x.device).zero_()
            )
            self.gamma = Parameter(
                self.gamma.new_empty(shape, dtype=x.dtype, device=x.device).zero_()
            )
            self.delta = Parameter(
                self.delta.new_empty(shape, dtype=x.dtype, device=x.device).zero_()
            )
            self.xi = Parameter(
                self.xi.new_empty(shape, dtype=x.dtype, device=x.device).zero_()
            )
            self.psi = Parameter(
                self.psi.new_empty(shape, dtype=x.dtype, device=x.device).zero_()
            )
            self.theta = Parameter(
                self.theta.new_empty(shape, dtype=x.dtype, device=x.device).zero_()
            )
            self.lam = Parameter(
                self.lam.new_empty(shape, dtype=x.dtype, device=x.device).zero_()
            )

        self._lazy_materialized = True

    def reset_parameters(self):
        params = (
            self.alpha,
            self.beta,
            self.gamma,
            self.delta,
            self.xi,
            self.psi,
            self.theta,
            self.lam,
        )
        for p in params:
            if not isinstance(p, UninitializedParameter):
                with torch.no_grad():
                    p.zero_()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if isinstance(self.alpha, UninitializedParameter):
            self._infer_parameters(x)

        if x.dim() != 4:
            raise ValueError(
                f"Expected 4D input (N, C, H, W), got {x.dim()}D with shape {tuple(x.shape)}"
            )
        if self.num_channels is not None and x.shape[1] != self.num_channels:
            raise RuntimeError(
                f"WideMELU was initialized with C={self.num_channels} but got C={x.shape[1]}."
                " Create a new WideMELU for different channel sizes."
            )

        X_norm = x / self.maxInput
        Y = torch.roll(X_norm, shifts=-1, dims=1)

        term1 = F.relu(X_norm)
        term2 = self.alpha * torch.clamp(X_norm, max=0)
        dist_sq_beta = (X_norm - 2) ** 2 + (Y - 2) ** 2
        dist_sq_gamma = (X_norm - 1) ** 2 + (Y - 1) ** 2
        dist_sq_delta = (X_norm - 1) ** 2 + (Y - 3) ** 2
        dist_sq_xi = (X_norm - 3) ** 2 + (Y - 1) ** 2
        dist_sq_psi = (X_norm - 3) ** 2 + (Y - 3) ** 2
        dist_sq_theta = (X_norm - 1) ** 2 + (Y - 2) ** 2
        dist_sq_lambda = (X_norm - 3) ** 2 + (Y - 2) ** 2
        term3 = self.beta * torch.sqrt(F.relu(2 - dist_sq_beta))
        term4 = self.gamma * torch.sqrt(F.relu(1 - dist_sq_gamma))
        term5 = self.delta * torch.sqrt(F.relu(1 - dist_sq_delta))
        term6 = self.xi * torch.sqrt(F.relu(1 - dist_sq_xi))
        term7 = self.psi * torch.sqrt(F.relu(1 - dist_sq_psi))
        term8 = self.theta * torch.sqrt(F.relu(1 - dist_sq_theta))
        term9 = self.lam * torch.sqrt(F.relu(1 - dist_sq_lambda))
        Z_norm = term1 + term2 + term3 + term4 + term5 + term6 + term7 + term8 + term9
        return Z_norm * self.maxInput



import torch
from torch.nn.parameter import Parameter, UninitializedParameter


class SReLU(torch.nn.modules.lazy.LazyModuleMixin, torch.nn.Module):
    def __init__(
        self,
        alpha_init: float = 0.0,
        beta_init: float = 0.0,
        gamma_init: float = 1.0,
        delta_init: float = 1.0,
    ):
        super().__init__()
        self.alpha_init_val = float(alpha_init)
        self.beta_init_val = float(beta_init)
        self.gamma_init_val = float(gamma_init)
        self.delta_init_val = float(delta_init)
        self.alpha: torch.Tensor = UninitializedParameter()
        self.beta: torch.Tensor = UninitializedParameter()
        self.gamma: torch.Tensor = UninitializedParameter()
        self.delta: torch.Tensor = UninitializedParameter()
        self.num_channels = None

    def _infer_parameters(self, x: torch.Tensor):
        if x.dim() < 2:
            raise ValueError(
                f"Input must have shape (N, C, ...). Got dim={x.dim()} with shape {tuple(x.shape)}"
            )

        c = x.shape[1]
        self.num_channels = int(c)

        param_shape = [1] * x.dim()
        param_shape[1] = c

        with torch.no_grad():
            self.alpha = Parameter(
                self.alpha.new_empty(param_shape, dtype=x.dtype, device=x.device).fill_(
                    self.alpha_init_val
                )
            )
            self.beta = Parameter(
                self.beta.new_empty(param_shape, dtype=x.dtype, device=x.device).fill_(
                    self.beta_init_val
                )
            )
            self.gamma = Parameter(
                self.gamma.new_empty(param_shape, dtype=x.dtype, device=x.device).fill_(
                    self.gamma_init_val
                )
            )
            self.delta = Parameter(
                self.delta.new_empty(param_shape, dtype=x.dtype, device=x.device).fill_(
                    self.delta_init_val
                )
            )

        self._lazy_materialized = True

    def reset_parameters(self):
        if not isinstance(self.alpha, UninitializedParameter):
            with torch.no_grad():
                self.alpha.fill_(self.alpha_init_val)
                self.beta.fill_(self.beta_init_val)
                self.gamma.fill_(self.gamma_init_val)
                self.delta.fill_(self.delta_init_val)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if isinstance(self.alpha, UninitializedParameter):
            self._infer_parameters(x)

        start = self.beta + self.alpha * (x - self.beta)
        finish = self.delta + self.gamma * (x - self.delta)
        out = torch.where(x < self.beta, start, x)
        out = torch.where(x > self.delta, finish, out)
        return out
